/**
 * JavaScript shell around the WAForth WebAssembly module.
 *
 * Provides higher-level functions to interact with the WAForth WebAssembly module.
 *
 * To the WebAssembly module, provides the infrastructure to dynamically load WebAssembly modules and
 * the I/O primitives with the UI.
 * */
declare class WAForth {
    #private;
    core?: WebAssembly.Instance;
    /**
     * Callback that is called when a character needs to be emitted.
     *
     * `c` is the single-character string that is emitted
     */
    onEmit?: (c: string) => void;
    key: () => number;
    constructor();
    /**
     * Initialize WAForth.
     *
     * Needs to be called before interpret().
     */
    load(): Promise<void>;
    memory(): WebAssembly.Memory;
    here(): number;
    pop(): number;
    popString(): string;
    push(n: number): void;
    pushString(s: string, offset?: number): number;
    /**
     * Read data `s` into the input buffer without interpreting it.
     */
    read(s: string): void;
    /**
     * Read data `s` into the input buffer, and start interpreter.
     */
    interpret(s: string, silent?: boolean): any;
    /**
     * Bind `name` to SCALL in Forth.
     *
     * When an SCALL is done with `name` on the top of the stack, `fn` will be called (with the name popped off the stack).
     * Use `stack` to pop parameters off the stack, and push results back on the stack.
     */
    bind(name: string, fn: (f: WAForth) => void): void;
    /**
     * Bind async `name` to SCALL in Forth.
     *
     * When an SCALL is done with `name` on the top of the stack, `fn` will be called (with the name popped off the stack).
     * Expects an execution token on the top of the stack, which will be called when the async callback is finished.
     * The execution parameter will be called with the success flag set.
     */
    bindAsync(name: string, fn: (f: WAForth) => Promise<void>): void;
}
export default WAForth;
